#!/usr/bin/env python3

import os
import logging
import sys
from elasticsearch import Elasticsearch
from elasticsearch.client import IndicesClient
import curator
from pscr_common.log import setup_logging
from pscr_common.health_check_metric import health_metric

_logger = logging.getLogger(__name__)

endpoint = os.environ.get('ES_ENDPOINT', None)
host = os.environ.get('ES_ENDPOINT', 'http://localhost:9200')
region = os.environ.get('REGION', 'us-east-1')
batch_size = int(os.environ.get('BATCH_SIZE', '100'))
user = os.environ.get('ES_USER', None)
password = os.environ.get('ES_PASSWORD', None)
svc_version=os.environ.get("SVC_VERSION", "UNKNOWN")
env_id=os.environ.get("ENV_ID", "UNKNOWN")

def main():
    global host

    setup_logging(level=os.environ.get("LOG_LEVEL", "INFO"), svc_version=svc_version, env_id=env_id)
    _logger.info("Starting aggregator...")
    # Mosaic 2.0 gives us a plain hostname, we need to turn it into an URL.
    if endpoint:
        host = f"https://{endpoint}"

    # Build the Elasticsearch client.
    if user and password:
        es = Elasticsearch(
            hosts=[host],
            verify_certs=True,
            http_auth=(user, password)
        )
    else:
        es = Elasticsearch(
            hosts=[host],
            verify_certs=True,
        )

    # Rollover index if necessary
    ic = IndicesClient(es)
    ic.rollover(alias="telemetry_index", body={"conditions" : {"max_age" : "1d"}})
    ic.rollover(alias="watchlist_telemetry_index", body={"conditions" : {"max_age" : "1d"}})

    # Filters by name and age, anything with telemetry_index- prefix and created more than 7 days ago
    # We put 8 here because an index created 7 days ago wil have data from 6-7 days ago, whereas an
    # index created 8 days ago will only have data 7-8 days old.
    index_list = curator.IndexList(es)
    index_list.filter_by_regex(kind='prefix', value='telemetry_index-')
    index_list.filter_by_age(source='creation_date', direction='older', unit='days', unit_count=8)

    _logger.info("Found %d report telemetry indices to delete", len(index_list.indices))

    # If our filtered list contains any indices, delete them.
    if index_list.indices:
        curator.DeleteIndices(index_list).do_action()

    # Filters by name and age, anything with telemetry_index- prefix and created more than 30 days ago
    # We put 31 here because an index created 30 days ago wil have data from 29-30 days ago, whereas an
    # index created 31 days ago will only have data 30-31 days old.
    index_list = curator.IndexList(es)
    index_list.filter_by_regex(kind='prefix', value='watchlist_telemetry_index-')
    index_list.filter_by_age(source='creation_date', direction='older', unit='days', unit_count=31)

    _logger.info("Found %d watchlist telemetry indices to delete", len(index_list.indices))

    # If our filtered list contains any indices, delete them.
    if index_list.indices:
        curator.DeleteIndices(index_list).do_action()

    # Now that old indexes have been pruned and current one rolled over if necessary, do aggregations on
    # remaining indices

    # Do the 1 day aggregations
    process_aggregations(es, 1, "global_hit_rate_1d")

    # Do the 7 day aggregations
    process_aggregations(es, 7, "global_hit_rate_1w")

    if os.environ.get("INFLUXDBREPORTER_ENABLED", None):
        health_metric(env_id, "feed-telemetry-aggregator")


def process_aggregations(es, days, telemetry_field):
    query = {
        "query" : {
            "range" : {
                "timestamp" : {
                    "gte" : "now-"+str(days)+"d",
                    "lt" : "now"
                }
            }
        },
        "size" : 0,
        "aggs" : {
            "my_buckets": {
                "composite" : {
                    "size" : batch_size,
                    "sources" : [
                        { "report_id": { "terms": {"field": "report_id" } } }
                    ]
                },
                "aggregations": {
                    "executions_sum": {
                        "sum": { "field": "executions" }
                    },
                    "hits_sum": {
                        "sum": { "field": "hits" }
                    }
                }
            }
        }
    }

    report_count = 0

    # Loop through grabbing only [composite.size] buckets at a time.
    while True:
        # Perform the search in elasticsearch
        res = es.search(index='telemetry_index-*', body=query)

        # Break if we didn't get anything back we are done
        if ('aggregations' not in res or
                'my_buckets' not in res['aggregations'] or
                'buckets' not in res['aggregations']['my_buckets']):
            break

        # Buckets will hold an array of our aggregations
        buckets = res['aggregations']['my_buckets']['buckets']
        if len(buckets) == 0:
            break

        # Process buckets
        updates = []
        for bucket in buckets:
            report_count += 1
            telemetry_value = round(100.0 * bucket['hits_sum']['value'] / max(bucket['executions_sum']['value'], 1), 7)
            updates.append("{\"update\" : {\"_id\" : \"" + bucket['key']['report_id'] + "\"}}")
            updates.append("{\"doc\" : {\"" + telemetry_field + "\" : "  + str(telemetry_value) + "}}")

        bulk_body = '\n'.join(updates) + '\n'
        result = es.bulk(index='report_index', doc_type='_doc', body=bulk_body)
        has_errors = result['errors']
        if has_errors:
            log_errors(result)

        # Grab the after key which is basically the cursor for the next loop iteration
        if 'after_key' not in res['aggregations']['my_buckets']:
            break
        after_key = res['aggregations']['my_buckets']['after_key']
        query['aggs']['my_buckets']['composite']['after'] = after_key

    _logger.info("Aggregated %d reports across %d days.", report_count, days)


def log_errors(result):
    for item in result['items']:
        if 'error' in item['update']:
            if item['update']['error']['type'] == 'document_missing_exception':
                _logger.info("Report with id %s was missing, could not update report_index.", item['update']['_id'])
            else:
                _logger.error("Error occurred during update of report_index: %s", str(item['update']['error']))

if __name__ == "__main__":
    sys.exit(main())

